/**
 * 
 */
package sim.listener;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import sim.exception.SNoImplementationException;
import sim.graphics.camera.SMovableCamera;
import sim.util.SLog;
import sim.util.SMouseEventHandler;

/**
 * La classe <b>SListenerCamera</b> est une camra qui implmente l'interface <b>MouseMotionListener</b> et <b>MouseListener</b>.
 * 
 * @author Simon Vezina
 * @since 2020-06-15
 * @version 2022-05-30
 */
public class SListenerCamera implements  MouseMotionListener, MouseListener {

  private static final boolean REVERSE_MODE = false;
  
  private static final double DEFAULT_MOVEMENT_SENSIBILITY = 0.1;
  private static final double DEFAULT_ROTATION_SENSIBILITY = 0.2;
  
  /**
   * La variable <b>mouse</b> reprsente les informations obtenues par la souris qui sera communique  la camra.
   */
  private final SMouseEventHandler mouse;
  
  /**
   * La variable <b>camera</b> reprsente la camra pouvant tre manipule.
   */
  private SMovableCamera camera;
  
  /**
   * La variable <b>movement_sensibility</b> reprsente le nombre d'unit de dplacement associ  chaque pixel de mouvement de la souris.
   */
  private double movement_sensibility;
  
  /**
   * La variable <b>rotation_sensibility</b> reprsente le nombre de degr de rotation associ  chaque pixel de mouvement de la souris.
   */
  private double rotation_sensibility;
  
  /**
   * 
   */
  public SListenerCamera(SMovableCamera camera) 
  {
    this.mouse = new SMouseEventHandler();
    this.camera = camera;
    
    this.movement_sensibility = DEFAULT_MOVEMENT_SENSIBILITY;
    this.rotation_sensibility = DEFAULT_ROTATION_SENSIBILITY;
  }

  /**
   * Mthode pour obtenir la camera du listener.
   * 
   * @return La camera
   */
  public SMovableCamera getCamera()
  {
    return this.camera;
  }
  
  /**
   * Mthode pour dfinir la camera du listener.
   * 
   * @param camera La camera.
   */
  public void setCamera(SMovableCamera camera)
  {
    this.camera = camera;
  }
  
  /**
   * Mthode pour dfinir la sensibilit du mouvement de la camra.
   * Cette valeur correspondra au dplacement unitaire de la camra par pixel de dplacement de la souris.
   * 
   * @param sensibility La sensibilit.
   */
  public void setMovementSensibility(double sensibility)
  {
    this.movement_sensibility = sensibility;
  }
  
  /**
   * Mthode pour dfinir la sensibilit de rotation de la camera.
   * Cette valeur correspondra au nombre de degr de rotation de la camra par pixel de dplacement de la souris.
   * 
   * @param sensibility La sensibilit.
   */
  public void setRotationSensibility(double sensibility)
  {
    this.rotation_sensibility = sensibility;
  }
  
  @Override
  public void mouseClicked(MouseEvent e) 
  {
    mouse.clicked(e);
  }

  @Override
  public void mousePressed(MouseEvent e) 
  {
    mouse.pressed(e);
  }

  @Override
  public void mouseReleased(MouseEvent e)
  {
    mouse.released(e);
  }

  @Override
  public void mouseEntered(MouseEvent e)
  {
        
  }

  @Override
  public void mouseExited(MouseEvent e)
  {
        
  }

  @Override
  public void mouseDragged(MouseEvent e)
  {
    mouse.dragged(e);
    
    // En raison d'un laboratoire, nous allons introduire un try/catch sur l'excution du mouvement.
    try {
      
      // Effectuer le dplacement de la camra en fonction du mode.
      if(REVERSE_MODE)
        reverseCameraMovement();
      else
        cameraMovement();
      
    }catch(SNoImplementationException exception) {
      SLog.logWriteLine("Message SListenerCamera --- Une opration sur la camra n'est pas admissible, car une mthode requise n'a pas t implmente.");
    }
  }

  @Override
  public void mouseMoved(MouseEvent e) 
  {
    mouse.moved(e);    
  }

  /**
   * Mthode pour raliser le mouvement de la camera en mode normal.
   * @param e L'vnement de souris.
   */
  protected void cameraMovement()
  {
    // BOUTON 1 : Enfonc.
    if(mouse.isButtonPressed(1))
    {
      // BOUTON 3 : Non enfonc 
      if(!mouse.isButtonPressed(3))
      {
        // Rotation Yaw (gauche-droite) 
        // Remarque : Lorque la souris effectue un dplacement vers la droite, la rotation est positif.
        camera.rotationYaw(mouse.getDeltaX()*rotation_sensibility);
        
        // Rotation Pitch(haut-bas).
        // Remarque : Lorque la souris effectue un dplacement vers le haut, la rotation est ngative.
        camera.rotationPitch(mouse.getDeltaY()*rotation_sensibility*-1.0);
      }
      else
      {
        // Rotation Roll (autour de l'axe avant).
        camera.rotationRoll(mouse.getDeltaX()*rotation_sensibility);
      }
    }else
    {
      // BOUTON 3 : Enfonc.
      if(mouse.isButtonPressed(3))
      {
        // Dplacement vers l'avant-arrire.
        // Remarque : Lorque la souris effectue un dplacement vers le haut, le dplacement est ngatif.
        camera.moveBack(mouse.getDeltaY()*movement_sensibility);
        
        // Dplacement vers la droite-gauche.
        // Remarque : Lorque la souris effectue un dplacement vers la droite, le dplacement est positive.
        camera.moveRight(mouse.getDeltaX()*movement_sensibility);
      }
    }
  }
  
  /**
   * Mthode pour raliser le mouvement de la camera en mode inverse (mode touchpad).
   * 
   * @param e L'vnement de souris.
   */
  protected void reverseCameraMovement()
  {
    // BOUTON 1 : Enfonc.
    if(mouse.isButtonPressed(1))
    {
      // BOUTON 3 : Non enfonc 
      if(!mouse.isButtonPressed(3))
      {
        // Rotation Yaw (gauche-droite) 
        // Remarque : Lorque la souris effectue un dplacement vers la droite, la rotation est positif.
        // On va inverser.
        camera.rotationYaw(mouse.getDeltaX()*rotation_sensibility*-1.0);
        
        // Rotation Pitch(haut-bas).
        // Remarque : Lorque la souris effectue un dplacement vers le haut, la rotation est ngative.
        // On va inverser.
        camera.rotationPitch(mouse.getDeltaY()*rotation_sensibility);
      }
      else
      {
        // Rotation Roll (autour de l'axe avant).
        camera.rotationRoll(mouse.getDeltaX()*rotation_sensibility);
      }
    }else
      // BOUTON 3 : Enfonc.
      if(mouse.isButtonPressed(3))
      {
        // Dplacement vers l'avant-arrire.
        // Remarque : Lorque la souris effectue un dplacement vers le haut, le dplacement est ngatif.
        // On va inverser.
        camera.moveFront(mouse.getDeltaY()*movement_sensibility);
        
        // Dplacement vers la droite-gauche.
        // Remarque : Lorque la souris effectue un dplacement vers la droite, le dplacement est positive.
        // On va inverser.
        camera.moveLeft(mouse.getDeltaX()*movement_sensibility);
      }
  }
  
}
