/**
 * 
 */
package sim.math;

import java.util.Arrays;

import sim.exception.SConstructorException;
import sim.util.SArrays;

/**
 * La classe <b>SQuaternion</b> reprsente un quaternion.
 * 
 * @author Simon Vezina
 * @since 2020-04-10
 * @version 2022-06-02 (version labo  Le champ lectrique v1.1.0)
 */
public class SQuaternion {

  /**
   * La constante <b>ZERO_ROTATION_AXIS</b> reprsente l'axe de rotation d'un quaternion sans rotation.
   */
  private static final SVector3d ZERO_ROTATION_AXIS = SVector3d.UNITARY_X;
  
  /**
   * La constante <b>ZERO</b> correspond  un quaternion nul o l'ensemble de ses composants sont zro (0.0, 0.0, 0.0, 0.0).
   */
  public static final SQuaternion ZERO = new SQuaternion(0.0, 0.0, 0.0, 0.0);
  
  /**
   * La constante <b>ZERO_ROTATION</b> correspond  un quaternion de rotation (module gale  1), mais dont la rotation est nulle.
   */
  public static final SQuaternion ZERO_ROTATION = new SQuaternion(0.0, ZERO_ROTATION_AXIS);
  
  /**
   * La variable <b>q</b> reprsente un tableau contenant les 4 composantes du quaternion quant q = q0 + qx*i + qy*j + qz*k = (q0, qx, qy, qz).
   */
  private final double[] q;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un quaternion.
   * 
   * @param q_w La composante constante du quaternion (w).
   * @param q_x La composante x du quaternion (x).
   * @param q_y La composante y du quaternion (y).
   * @param q_z La composante z du quaternion (z).
   */
  public SQuaternion(double q_w, double q_x, double q_y, double q_z)
  {
    this.q = new double[4];
    
    this.q[0] = q_w;
    this.q[1] = q_x;
    this.q[2] = q_y;
    this.q[3] = q_z;
  }
  
  /**
   * Constructeur d'une quaternion  partir d'un tableau de 4 donnes.
   * 
   * @param q Le tableau de 4 donnes
   * @throws SConstructorException Si le tableau ne contient pas exactement 4 donnes.
   */
  public SQuaternion(double... q) throws SConstructorException
  {
    if(q.length != 4)
      throw new SConstructorException("Erreur SQuaternion 001 : La taille du tableau dfinissant le quaternion est de " + q.length + " ce qui n'est pas gal  4.");
    
    this.q = new double[4];
    
    this.q[0] = q[0];
    this.q[1] = q[1];
    this.q[2] = q[2];
    this.q[3] = q[3];
  }
  
  /**
   * Constructeur d'un quaternion permettant de constuire un quaternion reprsentant une rotation 
   * d'un angle donn autour d'un axe donn. Pour que la construction du quaternion puisse corresponde
   *  une rotation, l'angle doit tre en radiant et l'axe de rotation doit tre unitaire.
   * 
   * @param angle L'angle de rotation <u>en radian</u>.
   * @param axis L'axe de la rotation sous la reprsentation d'un vecteur unitaire (longueur gal  1).
   * @throws SconstructorException Si l'axe de rotation n'est pas un vecteur unitaire.
   */
  public SQuaternion(double angle, SVector3d axis) throws SConstructorException
  {
    // Cas particulier : Aucune angle de rotation, alors dfinir un axe par dfaut.
    // Validation : Si aucun axe de rotation, alors le quaternion n'est pas valide.
    if(angle == 0.0)
      axis = ZERO_ROTATION_AXIS;
    else
      if(axis.isZero())
        throw new SConstructorException("Erreur SQuaternion 002 : Le quaternion ne peut pas tre dfini, car l'angle de rotation gal " + angle + ", mais l'axe de rotation est nul.");
        
    // Validation : Si l'axe de rotation n'est pas unitaire, alors le quaternion n'est pas valide.
    if(!axis.isUnitVector())
      throw new SConstructorException("Erreur SQuaternion 003 : Le quaternion ne peut pas reprsenter une rotation, car l'axe de rotation axis = " + axis + " n'est pas unitaire.");
          
    //  Si l'angle de rotation est ngatif, il faut inverser l'axe de rotation.
    if(angle < 0)
    {
      angle *= -1;
      axis = axis.multiply(-1.0);
    }
      
    // Allocation de la mmoire.
    this.q = new double[4];
    
    // Dfinir les compostantes du quaternion.
    double sin = Math.sin(angle/2.0);
    
    this.q[0] = Math.cos(angle/2.0);
    this.q[1] = sin * axis.getX();
    this.q[2] = sin * axis.getY();
    this.q[3] = sin * axis.getZ();
  }
  
  
   
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir un quaternion unitaire  partir d'un vecteur 3d contenant les trois composante (x, y, z) du quaternion.
   * Puisque le quaternion calcul sera unitaire, nous pouvons dduire la composante w de (w, x, y, z) afin de reconstruire le quaternion  partir de 3 composantes.
   *  
   * @param v Le vecteur utilis pour reconstruire le quaternion.
   * @return Le quaternion unitaire.
   * @throws Si le vecteur ne permet pas de gnrer un quaternion unitaire.
   */
  public static SQuaternion toUnitQuaternion(SVector3d v) throws IllegalArgumentException
  {
    double w_2 = 1.0 - v.getX()*v.getX() - v.getY()*v.getY() - v.getZ()*v.getZ();
    
    if(w_2 < 0.0)
      throw new IllegalArgumentException("Erreur SQuaternion 003 : Le vecteur v = " + v + " ne permet pas de gnrer un quaterion unitaire, car w*w = " + w_2 + " est ngatif.");
    
    return new SQuaternion(Math.sqrt(w_2), v.getX(), v.getY(), v.getZ());
  }
  
  /**
   * Mthode pour obtenir un quaternion  partir d'un vecteur 3d o la composante q_w du quaternion sera zro
   * et les composante q_x, q_y et q_z seront gales aux composante du vecteur. 
   * Ainsi, le quaternion construit sera q = (0, v_x, v_y, v_z).
   * Cette construction de quaternion permet d'obtenir un quaternion de type position.
   * 
   * @param v Le vecteur.
   * @return Le quaternion de type position.
   */
  public static SQuaternion toPositionQuaternion(SVector3d v)
  {
    return new SQuaternion(0.0, v.getX(), v.getY(), v.getZ());
  }
  
  /**
   * Mthode pour obtenir un quaternion de rotation  partir d'un angle de rotation et d'un axe de rotation.
   * Le vecteur reprsentant l'axe de rotation sera normalis dans le processus de cration du quaternion.
   * Seul l'angle en radians sera considr pour tablir l'angle de rotation du quaternion.
   * 
   * @param radians L'angle de rotation.
   * @param v L'axe de rotation.
   * @return Le quaternion de type rotation.
   */
  public static SQuaternion toRotationQuaternion(double radians, SVector3d v) throws IllegalArgumentException
  {
    // Cas particulier : Aucun axe de rotation.
    if(v.isZero())
      if(radians == 0.0)
        return new SQuaternion(radians, v);
      else
        throw new IllegalArgumentException("Erreur SQuaternion 004 : Le quaternion ne peut pas tre dfini, car l'angle de rotation gal " + radians + ", mais l'axe de rotation est nul.");
      
    // Cas gnral : Construire un quanternion avec un axe de rotation normalis.
    return new SQuaternion(radians, v.normalize());
  }
  
  /**
   * Mthode pour obtenir le quaternion de la rotation permettant d'orienter le vecteur itnial dans l'orientation du vecteur final.
   * 
   * @param initial_vector Le vecteur initial.
   * @param final_vector Le vecteur final.
   * @return Le quaternion de la rotation.
   */
  public static SQuaternion toRotationQuaternion(SVector3d initial_vector, SVector3d final_vector) throws IllegalArgumentException
  {
    // Vrification que l'on peut normaliser nos vecteurs.
    try {
      
      initial_vector = initial_vector.normalize();
      final_vector = final_vector.normalize();
    
    }catch(SImpossibleNormalizationException e) {
      throw new IllegalArgumentException(e);
    }
    
    // Test du produit scalaire.
    double dot = initial_vector.dot(final_vector);
    
    // Cas particulier : Est-ce que est orientation son parallle ou anti-parallle.
    if(SMath.nearlyEquals(dot, 1.0))
      return ZERO_ROTATION;
    else
      if(SMath.nearlyEquals(dot, -1.0))
      {
        // Sous cas particulier : Est-ce que le vecteur de rfrence est l'axe z, dont le vecteur final sera -z.
        if(initial_vector.equals(SVector3d.UNITARY_Z))
          return new SQuaternion(Math.PI, SVector3d.UNITARY_X);
        else
          throw new IllegalArgumentException("Ca l'aire qu'il faut grer un super cas de rotation de 180 degr");
      }
    
    // Dfinir l'angle avec le produit scalaire entre les deux vecteurs normalis.
    // Dfinir l'axe de rotation avec le produit vectoriel des deux vecteurs dont le rsultat sera normalis.
    return new SQuaternion(Math.acos(dot), initial_vector.cross(final_vector).normalize());    
  }
  
  /**
   * ...
   * 
   * @return
   */
  public double getW()
  {
    return q[0];
  }
  
  /**
   * ...
   * 
   * @return
   */
  public double getX()
  {
    return q[1];
  }
  
  /**
   * ...
   * 
   * @return
   */
  public double getY()
  {
    return q[2];
  }
  
  /**
   * ...
   * 
   * @return
   */
  public double getZ()
  {
    return q[3];
  }
  
  /**
   * Mthode pour obtenir la partie vectorielle du quaternion. Le vecteur correspond  (qx, qy, qz).
   * 
   * @return La partie vectorielle (qx, qy, qz) du quaternion (qw, qx, qy, qz).
   */
  public SVector3d getVector()
  {
    return new SVector3d(this.q[1], this.q[2], this.q[3]);
  }
  
  /**
   * Mthode pour valuer le module d'un quaternion.
   * 
   * @return Le module du quaternion.
   */
  public double modulus()
  {
    return Math.sqrt(this.q[0]*this.q[0] + this.q[1]*this.q[1] + this.q[2]*this.q[2] + this.q[3]*this.q[3]);
  }
  
  /**
   * Mthode pour normaliser le quaternion.
   * 
   * @return Le quaternion normalis.
   */
  public SQuaternion normalize()
  {
    double norm = this.modulus();
    
    return new SQuaternion(this.q[0] / norm, this.q[1] / norm, this.q[2] / norm, this.q[3] / norm);
  }
  
  /**
   * 
   * @param q
   * @return
   */
  public SQuaternion add(SQuaternion q)
  {
    SQuaternion empty = new SQuaternion();
    
    SArrays.add(this.q, q.q, empty.q);
    
    return empty;
  }
  
  /**
   * 
   * @param q
   * @return
   */
  public SQuaternion substract(SQuaternion q)
  {
    SQuaternion empty = new SQuaternion();
    
    SArrays.substract(this.q, q.q, empty.q);
    
    return empty;
  }
  
  /**
   * ...
   * 
   * @param s
   * @return
   */
  public SQuaternion multiply(double s)
  {
    SQuaternion empty = new SQuaternion();
    
    SArrays.multiply(this.q, s, empty.q);
    
    return empty;
  }
  
  /**
   * Mthode pour effectuer le produit de deux quaternion.
   * 
   * @param q Le quaternion  multiplier.
   * @return Le rsultat de la multiplication de deux quaternions.
   */
  public SQuaternion multiply(SQuaternion q)
  {
    double q0 = this.q[0]*q.q[0] - this.q[1]*q.q[1] - this.q[2]*q.q[2] - this.q[3]*q.q[3];
    double q1 = this.q[0]*q.q[1] + this.q[1]*q.q[0] + this.q[2]*q.q[3] - this.q[3]*q.q[2];
    double q2 = this.q[0]*q.q[2] - this.q[1]*q.q[3] + this.q[2]*q.q[0] + this.q[3]*q.q[1];
    double q3 = this.q[0]*q.q[3] + this.q[1]*q.q[2] - this.q[2]*q.q[1] + this.q[3]*q.q[0];
    
    return new SQuaternion(q0, q1, q2, q3);
  }
  
  /**
   * Mthode pour obtenir le conjug d'un quaternion.
   * 
   * @return
   */
  public SQuaternion conjugate()
  {
    return new SQuaternion(this.q[0], -1*this.q[1], -1*this.q[2], -1*this.q[3]);
  }
  
  /**
   * Mthode pour obtenir l'inverse d'un quaternion.
   * 
   * @return L'inverse du quaternion.
   */
  public SQuaternion inverse() throws IllegalArgumentException
  {
    double norm = this.modulus();
    
    if(SMath.nearlyZero(norm))
      throw new IllegalArgumentException("Erreur 004 : L'inversion du quaternion est impossible, car le module est nul.");
    
    double div = norm*norm;
    
    return this.conjugate().multiply(1.0 / div); 
  }
  
  
  
  /**
   * Mthode pour dterminer si le quaternion est normalis.
   * 
   * @return <b>true</b> si le quaternion est normalis et <b>false</b> sinon.
   */
  public boolean isNormalized()
  {
    return SMath.nearlyEquals(this.modulus(), 1.0);
  }
  
  /**
   * Mthode pour dterminer si le quaternion est de type <i>rotation</i>.
   * Un quaternion pourra reprsente une rotation (type rotation) si son module est gal  1.
   * 
   * @return <b>true</b> si le quaternion est de type rotation et <b>false</b> sinon.
   */
  public boolean isRotationQuaternion()
  {
    return this.isNormalized();
  }
  
  /**
   * Mthode pour dterminer si le quaternion est de type <i>position</i>.
   * Un quaternion pourra reprsenter une position (type position) si sa compostante q_w (composante constante) est gal  0.
   * 
   * @return <b>true</b> si le quaternion est de type position et <b>false</b> sinon.
   */
  public boolean isPositionQuaternion()
  {
    return SMath.nearlyZero(this.q[0]);
  }
  
  /**
   * Mthode pour obtenir l'angle de rotation associ au quaternion. Celui-ci doit tre normalis pour reprsenter une matrice de rotation.
   * 
   * @return L'angle de rotation <u>en radian</u> du quaternion.
   * @throws IllegalArgumentException Si le quaternion n'est pas normalis.
   */
  public double rotationAngle() throws IllegalArgumentException
  {
    if(!this.isNormalized())
      throw new IllegalArgumentException("Erreur SQuaternion 002 : L'angle de rotation ne peut pas tre dfini pour ce quaternion, car celui-ci n'est pas normalis.");
    
    // Dans le cas d'un quaternion normalis, nous avons q0 = cos(theta/2).
    // Nous allons prendre la valeur positive de l'arc cosius.
    return 2.0*Math.acos(this.q[0]);
  }
  
  /**
   * Mthode pour obtenir l'axe de rotation associ au quaternion. Celui-ci doit tre normalis pour reprsenter une matrice de rotation.
   * 
   * @return L'axe de rotation du quaternion (normalis).
   * @throws IllegalArgumentException Si le quaternion n'est pas normalis.
   */
  public SVector3d rotationAxis() throws IllegalArgumentException
  {
    if(!this.isNormalized())
      throw new IllegalArgumentException("Erreur SQuaternion 003 : L'axe de rotation ne peut pas tre dfini pour ce quaternion, car celui-ci n'a pas t normalis. Le module actuel est gal  " + this.modulus() + ".");
  
    // Obtenir l'angle de rotation en radian.
    double angle = this.rotationAngle();
    
    // Dans le cas d'un quaternion normalis, nous avons v = sin(theta/2)*u  o u est l'axe de rotation normalis et v = (qx, qy, qz).
    return this.getVector().multiply(1.0 / Math.sin(angle/2.0));
  }
  
  /**
   * Mthode pour appliquer une rotation sur un vecteur v.
   * 
   * @param v
   * @param q
   * @return
   * @throws IllegalArgumentException Si le quaternion n'est pas normalis.
   */
  public SVector3d rotation(SVector3d v) throws IllegalArgumentException
  {
    if(!this.isNormalized())
      throw new IllegalArgumentException("Erreur SQuaternion 004 : La rotation du vecteur ne peut pas tre ralise, car le quaternion n'a pas t normalis.");
  
    return this.multiply(SQuaternion.toPositionQuaternion(v)).multiply(this.conjugate()).getVector();
  }

 
  
  
  
  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + Arrays.hashCode(q);
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    SQuaternion other = (SQuaternion) obj;
    
    if (!SArrays.nearlyEquals(q, other.q))
      return false;
    
    return true;
  }

  @Override
  public String toString() 
  {
    return "[" + this.q[0] + ", " + this.q[1] + ", " + this.q[2] + ", " + this.q[3] +"]"; 
  }
  
}
