package sim.math;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.StringTokenizer;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe SVectorUV reprsentante une coordonne uv associe habituellement  une texture. 
 * Ce vecteur possde une composante u et v et sont <b>habituellement</b> situ dans l'intervalle [0.0 , 1.0].
 * Si ce n'est pas le cas, il est important de les recadrer les coordonnes uv  une valeur dans l'intervalle [0,1].
 * <p>Pour ce faire, il y a deux algorithmes possibles :
 * <ul>Repeat : Effectue une rptition de la coordonne de droite  gauche et de bas  haut (si dpassement positif).</ul>
 * <ul> Clamp : Coupe la coordonne aprs dpassement (u > 1 ==> u = 1, u < 0 ==> u = 0).</ul></p> 
 *  
 * @author Simon Vezina
 * @since 2015-09-15
 * @version 2022-04-19
 */
public class SVectorUV implements SVector {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante 'DEFAULT_COMPONENT' correspond  la coordonne par dfaut d'une composante u et v qui est gale  {@value}.
   */
  private static final double DEFAULT_COMPONENT = 0.0; 
  
  /**
   * La constante 'MIN_VALUE' correspond  la coordonne minimale d'une composante u et v qui est gale  {@value}.
   * <p>Cependant, une coordonne uv peut tre quand mme infrieure  {@value}, mais devra tre reconvertie  une valeur suprieure  {@value}
   * avant d'tre utilise pour chercher une information dans une texture. Si tel est le cas, le SVectorUV sera considr <b>out_of_bound</b>.</p>
   */
  private static final double MIN_VALUE = 0.0; // coordonne minimale
  
  /**
   * La constante 'MAX_VALUE' correspond  la coordonne maximale d'une composante u et v qui est gale  {@value}.
   * <p>Cependant, une coordonne uv peut tre quand mme suprieure  {@value}, mais devra tre reconvertie  une valeur infrieure  {@value}
   * avant d'tre utilise pour chercher une information dans une texture. Si tel est le cas, le SVectorUV sera considr <b>out_of_bound</b>.</p>
   */
  private static final double MAX_VALUE = 1.0; // coordonne maximale
  
  /**
   * La constante 'REPEAT' dfinit le choix de recadrer une coordonne uv  l'extrieur de l'intervalle [0,1]  l'aide d'une rptition des coordonnes uv.
   * <p>Par exemple : 
   * <ul>si u = 1.23, alors la coordonne recadre par rptition donnera u = 0.23.</ul>
   * <ul>si u = -2.8, alors la coordonne recadre par rptition donnera u = 0.2.</ul></p>
   */
  public static final int REPEAT = 0;
  
  /**
   * La constante 'CLAMP' dfinit le choix de recadrer une coordonne uv  l'extrieur de l'intervalle [0,1]  l'aide d'un arrt des coordonnes uv.
   * <p>Par exemple : 
   * <ul>si u = 1.23, alors la coordonne recadre par rptition donnera u = 1.0</ul>
   * <ul>si u = -2.8, alors la coordonne recadre par rptition donnera u = 0.0.</ul></p>
   */
  public static final int CLAMP = 1;
  
  //--------------
  // VARIABLES  //
  //--------------
  
  /**
   * La variable 'u' correspond  la composante u du vecteur uv. Cette coordonne est selon l'axe x (largeur, width) d'une texture o le (0,0) est situ dans le coin suprieur gauche.
   */
  private final double u;
  
  /**
   * La variable 'v' correspond  la composante v du vecteur uv. Cette coordonne est selon l'axe y (hauteur, height) d'une texture o le (0,0) est situ dans le coin suprieur gauche.
   */
  private final double v; 

  /**
   * La variable 'is_out_of_bound' dtermine si l'une des composantes uv du vecteur est  l'extrieur de l'intervalle [0,1]. 
   * Si tel est le cas, ces coordonnes devront tre recadres.  
   */
  private final boolean is_out_of_bound;
  
  //-----------------------------
  // CONSTRUCTEURS ET MTHODES //
  //-----------------------------
  
  /**
   * Constructeur reprsentant une coordonne uv  l'origine.
   */
  public SVectorUV()
  {
    this(DEFAULT_COMPONENT, DEFAULT_COMPONENT);
  }

  /**
   * Constructeur d'une coordonne uv quelconque.
   * 
   * @param u - La composante u de la coordonne uv.
   * @param v - La composante v de la coordonne uv.
   */
  public SVectorUV(double u, double v) throws SConstructorException
  {
    if (u < MIN_VALUE || u > MAX_VALUE || v < MIN_VALUE || v > MAX_VALUE)
      is_out_of_bound = true;
    else
      is_out_of_bound = false;
    
    this.u = u;
    this.v = v;
  }

  /**
   * Constructeur d'un vecteur uv en utilisant un string dcrivant les paramtres u et v du vecteur uv. 
   * Une lecture autorise peut prendre la forme suivante : "double u" "double v" (incluant la notation avec , ( ) [ ] comme dlimieur dans l'expression du string comme par exemple (2.3, 4.3) ).
   * 
   * @param string - Le string de l'expression du vecteur en paramtres u et v.
   * @throws SReadingException Si le format de la lecture n'est pas adquat.
   */
  public SVectorUV(String string)throws SConstructorException
  {
    try {
      double[] tab = read(string);
      
      //Vrification du respect de la contrainte 0 <= u,v <= 1
      if(tab[0] < MIN_VALUE || tab[0] > MAX_VALUE || tab[1] < MIN_VALUE || tab[1] > MAX_VALUE)
        is_out_of_bound = true;
      else
        is_out_of_bound = false;
      
      u = tab[0];
      v = tab[1];
      
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SVectorUV 001 : L'argument '" + string + "' n'est pas valide pour dfinir un vecteur uv.", e);
    }
  }
  
  /**
   * Mthode pour dterminer si la coordonne uv possde une composante dont la valeur est  l'extrieur de l'intervalle [0,1].
   * 
   * @return <b>true</b> s'il y a une composante  l'extrieur de l'intervalle et <b>false</b> sinon.
   */
  public boolean isOutOfBound()
  {
    return is_out_of_bound;
  }
  
  /**
   * Mthode pour obtenir la composante u de la coordonne uv.
   * 
   * @return La composante u de la coordonne uv.
   */
  public double getU()
  {
    return u;
  }

  /**
   * Mthode pour obtenir la composante v de la coordonne uv.
   * 
   * @return La composante v de la coordonne uv.
   */
  public double getV()
  {
    return v;
  }

  /**
   * Mthode qui permet d'obtenir un vecteur uv cadr dans l'intervalle [0,1]  l'aide de l'algoritme de rptition (REPEAT).
   *
   * @return Le vecteur uv recadr (si ncessaire) par l'algorithme de rptition.
   */
  public SVectorUV getInBound()
  {
    return getInBound(REPEAT);
  }
  
  
  /**
   * Mthode qui permet d'obtenir un vecteur uv cadr dans l'intervalle [0,1]  l'aide d'un algoritme particulier.
   * 
   * @param crop_code - Le code de l'algorithme
   * @return Le vecteur uv recadr (si ncessaire). 
   * @throws SRuntimeException Si le code de recadrement n'est pas reconnu par le systme.
   */
  public SVectorUV getInBound(int crop_code) throws SRuntimeException
  {
    if(!is_out_of_bound)
      return this;
    else
      switch(crop_code)
      {
        case REPEAT : return cropRepeat(); 
        
        case CLAMP : return cropClamp();
        
        default : throw new SRuntimeException("Erreur SVectorUV 001 : Le code de recadrement '" + crop_code + "' n'est pas reconnu par le systme.");
      }
  }
  
  /**
   * Mthode pour recadrer une coordonne uv par l'algoritme de rptition (REPEAT).
   * 
   * @return Le vecteur uv aprs recadrement par coupure (REPEAT).
   */
  private SVectorUV cropRepeat()
  {
    //la coordonne uv sorte de l'intervale [0,1] positivement ou ngativement. 
    //Nous allons rectifier la situation ... par rptition
    //Une coordonne suprieur  1 revient dans la texture par rptition par 0
    //Une coordonne intrieur  0 revient dans la texture par rptition par 1 
    double repeat_u = u;
    if(repeat_u > 0)
      while(repeat_u > 1.0)
        repeat_u -= 1.0;
    else
      while(repeat_u < 0)
        repeat_u += 1.0;
    
    double repeat_v = v;
    if(repeat_v > 0)
      while(repeat_v > 1.0)
        repeat_v -= 1.0;
    else
      while(repeat_v < 0)
        repeat_v += 1.0;
    
    return new SVectorUV(repeat_u, repeat_v);
  }
  
  /**
   * Mthode pour recadrer une coordonne uv par l'algoritme de coupure (CLAMP).
   * 
   * @return Le vecteur uv aprs recadrement par coupure (CLAMP).
   */
  private SVectorUV cropClamp()
  {
    double clamp_u = u;
    
    if(clamp_u < 0)
      clamp_u = 0.0;
    else
      if(clamp_u > 1)
        clamp_u = 1.0;
    
    double clamp_v = v;
    
    if(clamp_v < 0)
      clamp_v = 0.0;
    else
      if(clamp_v > 1)
        clamp_v = 1.0;
    
    return new SVectorUV(clamp_u, clamp_v);
  }
  
  @Override
  public SVector add(SVector v)
  {
    return add((SVectorUV) v);
  }
  
  /**
   * Mthode pour effectuer l'addition avec une coordonne uv.
   * 
   * @param uv - La coordonne uv  additionner.
   * @return La coordonne uv rsultant de l'addition.
   */
  public SVectorUV add(SVectorUV uv)
  {
    return new SVectorUV(u + uv.u, v + uv.v);
  }
  
  /**
   * Mthode pour effectuer la multiplication par un scalaire d'une constante avec la coordonne uv lanant l'appel de la mthode.
   * 
   * @param cst La constante  multiplier.
   * @return La coordonne uv rsultant de la multiplication par un scalaire.
   */
  public SVectorUV multiply(double cst)
  {
    return new SVectorUV(u*cst, v*cst);
  }

  @Override
  public double dot(SVector vector)
  {
    return dot((SVectorUV)vector);
  }
  
  /**
   * Mthode pour effectuer le <b>produit scalaire</b> entre deux vecteurs.
   * 
   * @param vector Le vecteur  mettre en produit scalaire.
   * @return Le produit scalaire entre les deux vecteurs.
   */
  public double dot(SVectorUV vector)
  {
    return (u*vector.u + v*vector.v);
  }
  
  @Override
  public String toString()
  {
    return "[" + u + ", " + v + "]";    
  }

  @Override
  public void write(BufferedWriter bw) throws IOException
  {
    bw.write(toString());
  }
  
  /**
   * Mthode utilisant un string comme paramtre pour dfinir les composantes u et v vecteur uv. 
   * Une lecture autorise peut prendre la forme suivante : "double u" "double v" 
   *  (en incluant la notation avec , ( ) [ ] < > comme dlimieur dans l'expression du string comme par exemple (2.3, 4.3) ).
   *  
   * @param string - Le string de l'expression du vecteur en paramtres u et v.
   * @return un tableau de deux lments tel que u = [0] et v = [1]. 
   * @throws IllegalArgumentException Si le format de lecture n'est pas adquat.
   */
  private double[] read(String string) throws IllegalArgumentException
  {
    StringTokenizer tokens = new StringTokenizer(string, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() != 2)
      throw new IllegalArgumentException("Erreur SVectorUV 003 : L'expression '" + string + "' ne contient pas exactement 2 paramtres pour les composantes uv du vecteur uv.");
    else
    {
      String s = "";          //String  convertir en double pour les composantes u et v.
      String comp = "";       //Nom de la composante en lecture utilise pour l'envoie d'une Exception s'il y a lieu.
      
      try
      {
        double[] tab = new double[2]; //Tableau des 2 composantes
        
        comp = "u";
        s = tokens.nextToken();
        tab[0] = Double.valueOf(s);
        
        comp = "v";
        s = tokens.nextToken();
        tab[1] = Double.valueOf(s);
        
        return tab;
        
      }catch(NumberFormatException e){ 
        throw new IllegalArgumentException("Erreur SVectorUV 004 : L'expression '" + s +"' n'est pas valide pour dfinir la composante '" + comp + "' du vecteur en cours.", e);
      }
    } 
  }
  
  @Override
  public int hashCode()
  {
    final int prime = 31;
    int result = 1;
    result = prime * result + (is_out_of_bound ? 1231 : 1237);
    long temp;
    temp = Double.doubleToLongBits(u);
    result = prime * result + (int) (temp ^ (temp >>> 32));
    temp = Double.doubleToLongBits(v);
    result = prime * result + (int) (temp ^ (temp >>> 32));
    return result;
  }

  /* (non-Javadoc)
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj) {
      return true;
    }
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof SVectorUV)) {
      return false;
    }
    SVectorUV other = (SVectorUV) obj;
    
    if(!SMath.nearlyEquals(u, other.u))
      return false;

    if(!SMath.nearlyEquals(v, other.v))
      return false;
    
    return true;
  }

}// fin de la classe SVectorUV
