/**
 * 
 */
package sim.physics;

import sim.exception.SNoImplementationException;
import sim.math.SImpossibleNormalizationException;
import sim.math.SMath;
import sim.math.SVector3d;
import sim.math.field.SUndefinedFieldException;

/**
 * La classe <b>SElectrostatics</b> permet d'effectuer des calculs en physique en lien avec l'lectrostatique.
 * 
 * @author Simon Vzina
 * @since 2017-05-30
 * @version 2022-06-02 (version labo  Le champ lectrique v1.1.0)
 */
public class SElectrostatics {

  /**
   * La constante <b>k</b> reprsente la <b>constante de Coulomb</b> tant gale  <b>{@value} N.m^2.C^-2</b>.
   */
  public static final double k = SPhysics.k;
  
  /**
   * La constante <b>EPSILON_ZERO</b> reprsente la <b>constante lectrique</b> tant gale  {@value}.
   */
  public static final double EPSILON_ZERO = SPhysics.EPSILON_ZERO;
  
  /**
   * La constante <b>NO_ELECTRIC_FIELD</b> reprsente une composante d'un champ lectrique tant gal  zro.
   */
  public static final SVector3d NO_ELECTRIC_FIELD = SVector3d.ZERO;
  
  /**
   * Mthode permettant d'valuer la loi de Coulomb entre deux charges lectriques.
   * 
   * @param Q La charge qui applique la force lectrique.
   * @param r_Q La position de la charge Q.
   * @param q La charge qui subit la force lectrique.
   * @param r_q La position de la charge q.
   * @return La force lectrique applique par la charge Q sur la charge q.
   */
  public static SVector3d coulombLaw(double Q, SVector3d r_Q, double q, SVector3d r_q)
  {
    
    // Vecteur dplacement entre Q et q.
    SVector3d r_Qq = r_q.substract(r_Q);
    
    // valuer le dplacement au cube.
    double m = r_Qq.modulus();
    double r3 = m*m*m;
    
    // Retourner la force lectrique.
    return r_Qq.multiply(k*Q*q/r3);
  }
  
  
  /**
   * Mthode permettant d'valuer la force lectrique applique sur une charge q plonge dans un champ lectrique E.
   * 
   * @param q La charge qui subit la force lectrique en coulombs (C).
   * @param E Le champ lectrique qui applique la force lectrique (N/C).
   * @return La force lectrique applique sur la charge (N).
   */
  public static SVector3d electricForce(double q, SVector3d E)
  {
    
    return E.multiply(q);
  }
  
  /**
   * Mthode pour obtenir le champ lectrique gnr par une particule ponctuelle.
   * 
   * @param r_p La position de la particule.
   * @param Q La charge de la particule.
   * @param r La position o le champ lectrique est valu.
   * @return Le champ lectrique.
   */
  public static SVector3d particleElectricField(SVector3d r_p, double Q, SVector3d r)
  {
    
    // valuer la distance entre les deux points.
    SVector3d r_Q = r.substract(r_p);
    double m = r_Q.modulus();
    
    // Cas particulier : Calcul du champ lectrique sur une particule.
    if(SMath.nearlyZero(m))
      return NO_ELECTRIC_FIELD;
    
    // Cas gnral :
    double r3 = m*m*m;
    
    return r_Q.multiply(k*Q/r3);
  }
  
  /**
   * Mthode permettant d'valuer le champ lectrique gnr par une sphre uniformment charge.
   * 
   * @param r_S La position de la sphre en mtres (m).
   * @param R Le rayon de la sphre en mtres (m).
   * @oaram Q La charge de la sphre en coulombs (C).
   * @param r La position o le champ lectrique est valu en mtre (m).
   * @return Le champ lectrique gnr par la sphre (N/C).
   */
  public static SVector3d sphereElectricField(SVector3d r_S, double R, double Q, SVector3d r) 
  {
    throw new SNoImplementationException();
   
  }
  
  /**
   * Mthode permettant d'valuer le champ lectrique gnr par une tige infinie uniformment charge (TRIUC).
   * 
   * @param r_R La position d'un point appartenant  la tige en mtres (m).
   * @param axis L'axe de la tige en mtres (m).
   * @param lambda La densit de charge de la tige (C/m).
   * @param r La position o le champ lectrique est valu en mtres (m).
   * @return Le champ lectrique gnr par la tige infinie (N/C).
   * @throws SImpossibleNormalizationException Si l'axe de la tige ne peut par tre normalis.
   */
  public static SVector3d infiniteRodElectricField(SVector3d r_R, SVector3d axis, double lambda, SVector3d r) throws SImpossibleNormalizationException
  {
    // Faire la normalisation de l'axe de la tige.
    axis = axis.normalize();
    
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode pour valuer le champ lectrique gnr par une tige uniformment charge <u>hors axe</u>.
   * 
   * @param r_A La position de l'extrmit A de la tige.
   * @param r_B La position de l'extrmit B de la tige.
   * @param lambda La densit de charge sur la tige.
   * @param r La position o est valu le champ lectrique.
   * @return Le champ lectrique.
   * @throws SImpossibleNormalizationException Si un vecteur ne peut pas tre normalis.
   * @throws SUndefinedFieldException Si le champ lectrique n'est pas dfini.
   */
  public static SVector3d finiteRodElectricFieldOutsideAxis(SVector3d r_A, SVector3d r_B, double lambda, SVector3d r) throws SImpossibleNormalizationException, SUndefinedFieldException
  {
    throw new SNoImplementationException();
    
  }
    
  /**
   * Mthode pour obtenir le champ lectrique gnr par une tige uniformment charge (sur l'axe ou hors axe).
   * 
   * @param r_A La position de l'extrmit A de la tige.
   * @param r_B La position de l'extrmit B de la tige.
   * @param Q La charge sur la tige.
   * @param r La position o est valu le champ lectrique.
   * @return Le champ lectrique.
   */
  public static SVector3d finiteRodElectricField(SVector3d r_A, SVector3d r_B, double Q, SVector3d r) 
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode permettant d'valuer le champ lectrique gnr par une plaque infinie uniformment charge (PPIUC).
   * 
   * @param r_P La position d'un point appartenant  la plaque en mtres (m).
   * @param n_P La normale  la surface de la plaque en mtres (m).
   * @param sigma La densit de charge surfacique de la plaque (C/m^2).
   * @param r Le champ lectrique gnr par la plaque infinie en mtres (m).
   * @return Le champ lectrique gnr par la plaque infinie (N/C).
   * @throws SImpossibleNormalizationException Si la normale au plan de la plaque ne peut par tre normalise.
   */
  public static SVector3d infinitePlateElectricField(SVector3d r_P, SVector3d n_P, double sigma, SVector3d r) throws SImpossibleNormalizationException
  {
    // Faire la normalisation de la normale  la surface de la plaque.
    n_P = n_P.normalize();
    
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode pour obtenir le potentiel lectrique gnr par une particule ponctuelle.
   * 
   * @param r_p La position de la particule.
   * @param Q La charge lectrique de la particule.
   * @param r La position o sera valu le potentiel lectrique.
   * @return Le potentiel lectrique.
   */
  public static double particleElectricPotential(SVector3d r_p, double Q, SVector3d r)
  {
    throw new SNoImplementationException();
    
  }
  
}
