package sim.util;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;

import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;
import sim.math.SMath;

/**
 * La classe <b>SArrays</b> permet d'effectuer des oprations sur des tableaux (Array) de type primitif.
 * 
 * @author Simon Vzina
 * @since 2017-12-15
 * @version 2022-06-02 (version labo  Le champ lectrique v1.1.0)
 */
public final class SArrays {

  /**
   * La constante <b>EPSILON</b> reprsente un nombre trs petit positif, mais non nul.
   * Il est gal  la valeur {@value} .
   */
  private static final double EPSILON = SMath.EPSILON;
  
  /**
   * <p>
   * Mthode pour dterminer si deux tableau de nombres de type double sont <b>relativement gaux</b>. 
   * En utilisant une approche de calcul de diffrence, on vrifie si pour chaque lment du tableau
   * <ul>a - b < EPSILON*ref</ul>  
   * afin de <b>valid l'galit</b> entre a et b (a == b). 
   * </p>
   * <p>
   * EPSILON est un seuil de prcision (voir EPSILON, mais traditionnelement 1e-10) et ref est une base de rfrence (la valeur absolue la plus leve parmis a et b).
   * Cependant, si les deux chiffres sont infrieurs  EPSILON, ils seront considrs comme gaux.
   * </p>
   * 
   * @param tab1 Le premier tableau  comparer.
   * @param tab2 Le deuxime tableau  comparer.
   * @return <b>true</b> si les deux tableaux sont <b>relativement gaux</b> et <b>false</b> sinon.
   */
  public static boolean nearlyEquals(double[] tab1, double[] tab2)
  {
    return nearlyEquals(tab1, tab2, SArrays.EPSILON);
  }
  
  /**
   * <p>
   * Mthode pour dterminer si deux tableau de nombres de type double sont <b>relativement gaux</b>. 
   * En utilisant une approche de calcul de diffrence, on vrifie si pour chaque lment du tableau
   * <ul>a - b < EPSILON*ref</ul>  
   * afin de <b>valid l'galit</b> entre a et b (a == b). 
   * </p>
   * <p>
   * EPSILON est un seuil de prcision et ref est une base de rfrence (la valeur absolue la plus leve parmis a et b).
   * Cependant, si les deux chiffres sont infrieurs  EPSILON, ils seront considrs comme gaux.
   * </p>
   * 
   * @param tab1 Le premier tableau  comparer.
   * @param tab2 Le deuxime tableau  comparer.
   * @param epsilon La prcision acceptable.
   * @return <b>true</b> si les deux tableaux sont <b>relativement gaux</b> et <b>false</b> sinon.
   */
  public static boolean nearlyEquals(double[] tab1, double[] tab2, double epsilon) 
  {
    // Vrifier que les deux tableaux ont la mme taille.
    if(tab1.length != tab2.length)
      return false;
    
    // Vrifier que l'ensemble des donne indexe par indexe sont "nearlyEquals".
    for(int i = 0; i < tab1.length; i++)
      if(!SMath.nearlyEquals(tab1[i], tab2[i], epsilon))
        return false;
    
    // Puisque l'ensemble des valeurs indexe par indexe sont "nearlyEquals", le tableau doit l'tre. 
    return true;
  }
  
  /**
   * Mthode pour effectuer une copie d'un tableau.
   * 
   * @param tab Le tableau  copier.
   * @param copy Le tableau qui va recevoir la copie.
   * @return Le tableau qui a reu la copie. 
   * @throws IllegalArgumentException Si les deux tableaux n'ont pas la mme longueur.
   */
  public static int[] copy(int[] tab, int[] copy) throws IllegalArgumentException
  {
    if(tab.length != copy.length)
      throw new IllegalArgumentException("Erreur SArrays 001 : Les deux tableaux n'ont pas la mme longueur (" + tab.length + " et " + copy + ").");
    
    System.arraycopy(tab, 0, copy, 0, tab.length);
    
    return copy;
  }
  
  /**
   * Mthode pour effectuer une copie d'un tableau.
   * 
   * @param tab Le tableau  copier.
   * @param copy Le tableau qui va recevoir la copie.
   * @throws IllegalArgumentException Si les deux tableaux n'ont pas la mme longueur.
   */
  public static void copy(int[] tab, byte[] copy) throws IllegalArgumentException
  {
    if(tab.length != copy.length)
      throw new IllegalArgumentException("Erreur SArrays 002 : Les deux tableaux n'ont pas la mme longueur (" + tab.length + " et " + copy + ").");
    
    System.arraycopy(tab, 0, copy, 0, tab.length);
  }
  
  /**
   * Mthode pour effectuer une copie d'un tableau.
   * 
   * @param tab Le tableau  copier.
   * @param copy Le tableau qui va recevoir la copie.
   * @throws IllegalArgumentException Si les deux tableaux n'ont pas la mme longueur.
   */
  public static void copy(byte[] tab, byte[] copy) throws IllegalArgumentException
  {
    if(tab.length != copy.length)
      throw new IllegalArgumentException("Erreur SArrays 003 : Les deux tableaux n'ont pas la mme longueur (" + tab.length + " et " + copy + ").");
    
    System.arraycopy(tab, 0, copy, 0, tab.length);
  }
  
  /**
   * Mthode pour effectuer une copie d'un tableau.
   * 
   * @param tab Le tableau  copier.
   * @param copy Le tableau qui va recevoir la copie.
   * @throws SRuntimeException Si les deux tableaux n'ont pas la mme longueur.
   */
  public static void copy(double[] tab, double[] copy) throws ArrayIndexOutOfBoundsException
  {
    if(tab.length != copy.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 004 : Les deux tableaux n'ont pas la mme longueur (" + tab.length + " et " + copy + ").");
    
    System.arraycopy(tab, 0, copy, 0, tab.length);
  }
  
  /**
   * Mthode pour mettre  zro l'ensemble des donnes d'un tableau.
   * 
   * @param array Le tableau.
   * @return Une rfrence sur le tableau pass en paramtre.
   */
  public static double[] clear(double[] array)
  {
    for(int i = 0; i < array.length; i++)
      array[i] = 0.0;
    
    return array;
  }
  
  /**
   * Mthode pour effectuer la concatnation de deux tableaux.
   * 
   * </u>Reference :</u> https://stackoverflow.com/questions/80476/how-can-i-concatenate-two-arrays-in-java 
   * @param first Le premier tableau.
   * @param second Le deuxieme tableau.
   * @return Le tableau concaten.
   */
  public static <T> T[] concatenate(T[] first, T[] second) 
  {
    T[] result = Arrays.copyOf(first, first.length + second.length);
    System.arraycopy(second, 0, result, first.length, second.length);
    
    return result;
  }
  
  /**
   * Mthode pour effectuer la concatnation de deux tableaux.
   * 
   * </u>Reference :</u> https://stackoverflow.com/questions/80476/how-can-i-concatenate-two-arrays-in-java 
   * @param first Le premier tableau.
   * @param second Le deuxieme tableau.
   * @return Le tableau concaten.
   */
  public static double[] concatenate(double[] first, double[] second) 
  {
    double[] result = Arrays.copyOf(first, first.length + second.length);
    System.arraycopy(second, 0, result, first.length, second.length);
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>l'addition</u> d'une valeur  un tableau.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab Le 1er tableau.
   * @param a La valeur  ajouter.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] add(double[] tab, double a) 
  {
    double[] result = new double[tab.length];
    
    for(int i = 0; i < tab.length; i++)
      result[i] = tab[i] + a;
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>l'addition</u> entre deux tableaux de mme dimension.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab1 Le 1er tableau.
   * @param tab2 Le 2e tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   * @throws ArrayIndexOutOfBoundsException Si les deux tableaux n'ont pas la mme taille.
   */
  public static double[] add(double[] tab1, double[] tab2) throws ArrayIndexOutOfBoundsException
  {
    // Vrification de la taille des tableaux.
    if(tab1.length != tab2.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 005 : Les deux tableaux n'ont pas la mme longueur (" + tab1.length + " et " + tab2.length + ").");
    
    //------------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //------------------------------------------------
    //throw new SNoImplementationException();
    
    return add(tab1, tab2, new double[tab1.length]);
  }
  
  /**
   * Mthode pour raliser <u>l'addition</u> entre deux tableaux de mme dimension.
   * Le rsultat de l'addition sera enregistr dans le 3e tableau pass en paramtre de la mthode.
   * 
   * @param tab1 Le 1er tableau.
   * @param tab2 Le 2e tableau.
   * @param result Le tableau contenant le rsultat.
   * @return Le tableau contenant le rsultat.
   * @throws ArrayIndexOutOfBoundsException Si les 3 tableaux ne sont pas de mme taille.
   */
  public static double[] add(double[] tab1, double[] tab2, double[] result) throws ArrayIndexOutOfBoundsException
  {
    //-------------------------------------------------------------------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //
    // P.S. Ce code n'est pas pour le labo fonction discrtes, mais ne peut pas tre disponible pour ce labo.
    //       l'origine, ceci devrait servir de tche pour le labo Apprentissage par rseau de neurones, 
    //      mais l'ide a t abandonn. Cependant, le code est exploit.
    //-------------------------------------------------------------------------------------------------------
    //throw new SNoImplementationException();
    
    // Vrification de la taille des tableaux.
    if(tab1.length != tab2.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 006 : Le tableau 1 et 2 n'ont pas la mme longueur (" + tab1.length + " et " + tab2.length + ").");
    
    if(tab2.length != result.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 007 : Le tableau 2 et 3 n'ont pas la mme longueur (" + tab2.length + " et " + result.length + ").");
    
    for(int i = 0; i < tab1.length; i++)
      result[i] = tab1[i] + tab2[i];
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>la soustraction</u> d'une valeur  un tableau.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab Le 1er tableau.
   * @param a La valeur  soustraire.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] substract(double[] tab, double a) 
  {
    return add(tab, -1*a);
  }
  
  /**
   * Mthode pour raliser <u>la soustraction</u> entre deux tableaux de mme dimension.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab1 Le 1er tableau.
   * @param tab2 Le 2e tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   * @throws ArrayIndexOutOfBoundsException Si les deux tableaux n'ont pas la mme taille.
   */
  public static double[] substract(double[] tab1, double[] tab2) throws ArrayIndexOutOfBoundsException
  {
    if(tab1.length != tab2.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 008 : Les deux tableaux n'ont pas la mme longueur (" + tab1.length + " et " + tab2.length + ").");
    
    //----------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //------------------------------------------------
    //throw new SNoImplementationException();
    
    return substract(tab1, tab2, new double[tab1.length]);
  }
  
  /**
   * Mthode pour raliser la <u>soustraction</u> entre deux tableaux de mme dimension.
   * Le rsultat de la soustraction sera enregistre dans le 3e tableau pass en paramtre de la mthode.
   * 
   * @param tab1 Le 1er tableau.
   * @param tab2 Le 2e tableau.
   * @param result Le tableau contenant le rsultat.
   * @return Le tableau contenant le rsultat.
   * @throws ArrayIndexOutOfBoundsException Si les 3 tableaux ne sont pas de mme taille.
   */
  public static double[] substract(double[] tab1, double[] tab2, double[] result) throws ArrayIndexOutOfBoundsException
  {
    //-------------------------------------------------------------------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //
    // P.S. Ce code n'est pas pour le labo fonction discrtes, mais ne peut pas tre disponible pour ce labo.
    //       l'origine, ceci devrait servir de tche pour le labo Apprentissage par rseau de neurones, 
    //      mais l'ide a t abandonn. Cependant, le code est exploit.
    //-------------------------------------------------------------------------------------------------------
    //throw new SNoImplementationException();
    
    // Vrification de la taille des tableaux.
    if(tab1.length != tab2.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 009 : Le tableau 1 et 2 n'ont pas la mme longueur (" + tab1.length + " et " + tab2.length + ").");
    
    if(tab2.length != result.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 010 : Le tableau 2 et 3 n'ont pas la mme longueur (" + tab2.length + " et " + result.length + ").");
    
    for(int i = 0; i < tab1.length; i++)
      result[i] = tab1[i] - tab2[i];
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>la multiplication</u> d'un tableau par un scalaire.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab Le 1er tableau.
   * @param a La valeur  multiplier.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] multiply(double[] tab, double a) 
  {
    //----------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //------------------------------------------------
    //throw new SNoImplementationException();
    
    return multiply(tab, a,  new double[tab.length]);
  }
  
  /**
   * Mthode pour raliser <u>la multiplication</u> d'un tableau par un scalaire.
   * Le rsultat de la multiplication sera enregistr dans le 2e tableau pass en paramtre.
   * 
   * @param tab Le tableau.
   * @param a Le scalaire.
   * @param result Le rsultat de la multiplication du tableau par un scalaire.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] multiply(double[] tab, double a, double[] result) throws ArrayIndexOutOfBoundsException
  {
    if(tab.length != result.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 009 : Les deux tableaux n'ont pas la mme longueur (" + tab.length + " et " + result.length + ").");
    
    //-------------------------------------------------------------------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //
    // P.S. Ce code n'est pas pour le labo fonction discrtes, mais ne peut pas tre disponible pour ce labo.
    //       l'origine, ceci devrait servir de tche pour le labo Apprentissage par rseau de neurones, 
    //      mais l'ide a t abandonn. Cependant, le code est exploit.
    //-------------------------------------------------------------------------------------------------------
    //throw new SNoImplementationException();
    
    for(int i = 0; i < tab.length; i++)
      result[i] = a*tab[i];
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>la multiplication</u> entre deux tableaux de mme dimension.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab1 Le 1er tableau.
   * @param tab2 Le 2e tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   * @throws ArrayIndexOutOfBoundsException Si les deux tableaux n'ont pas la mme taille.
   */
  public static double[] multiply(double[] tab1, double[] tab2) throws ArrayIndexOutOfBoundsException
  {
    // Vrification de la taille des tableaux.
    if(tab1.length != tab2.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 010 : Les deux tableaux n'ont pas la mme longueur (" + tab1.length + " et " + tab2.length + ").");
    
    //----------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //------------------------------------------------
    //throw new SNoImplementationException();
    
    return multiply(tab1, tab2, new double[tab1.length]);
  }
  
  /**
   * Mthode pour raliser <u>la multiplication</u> entre deux tableaux de mme dimension.
   * Le rsultat de la multiplication sera enregistr dans le 3e tableau pass en paramtre de la mthode.
   * 
   * @param tab1 Le 1er tableau.
   * @param tab2 Le 2e tableau.
   * @param result Le tableau contenant le rsultat.
   * @return Le tableau contenant le rsultat.
   * @throws ArrayIndexOutOfBoundsException Si les 3 tableaux ne sont pas de mme taille.
   */
  public static double[] multiply(double[] tab1, double[] tab2, double[] result) throws ArrayIndexOutOfBoundsException
  {
    // Vrification de la taille des tableaux.
    if(tab1.length != tab2.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 011 : Le tableau 1 et 2 n'ont pas la mme longueur (" + tab1.length + " et " + tab2.length + ").");
    
    if(tab2.length != result.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 012 : Le tableau 2 et 3 n'ont pas la mme longueur (" + tab2.length + " et " + result.length + ").");
   
    //-------------------------------------------------------------------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //
    // P.S. Ce code n'est pas pour le labo fonction discrtes, mais ne peut pas tre disponible pour ce labo.
    //       l'origine, ceci devrait servir de tche pour le labo Apprentissage par rseau de neurones, 
    //      mais l'ide a t abandonn. Cependant, le code est exploit.
    //-------------------------------------------------------------------------------------------------------
    //throw new SNoImplementationException();
    
    for(int i = 0; i < tab1.length; i++)
      result[i] = tab1[i] * tab2[i];
    
    return result;
  }
  /**
   * Mthode pour raliser <u>le produit scalaire</u> entre deux tableaux de mme dimension.
   * 
   * @param A Le 1er tableau.
   * @param B Le 2e tableau.
   * @return Le rsultat de l'opration mathmatique.
   * @throws ArrayIndexOutOfBoundsException Si les deux tableaux n'ont pas la mme taille.
   */
  public static double dot(double[] A, double[] B) throws ArrayIndexOutOfBoundsException
  {
    if(A.length != B.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 013 : Les deux tableaux A et B n'ont pas la mme longueur (A = " + A.length + " et B = " + B.length + ").");
       
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode pour raliser <u>la division</u> entre deux tableaux de mme dimension.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab1 Le 1er tableau.
   * @param tab2 Le 2e tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   * @throws ArrayIndexOutOfBoundsException Si les deux tableaux n'ont pas la mme taille.
   */
  public static double[] divide(double[] tab1, double[] tab2) throws ArrayIndexOutOfBoundsException
  {
    if(tab1.length != tab2.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 014 : Les deux tableaux n'ont pas la mme longueur (" + tab1.length + " et " + tab2 + ").");
    
    //----------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //------------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    double[] result = new double[tab1.length];
    
    for(int i = 0; i < tab1.length; i++)
      result[i] = tab1[i] / tab2[i];
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>l'exponentiel</u> d'un tableau.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab Le tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] exp(double[] tab)
  {
    double[] result = new double[tab.length];
    
    for(int i = 0; i < tab.length; i++)
      result[i] = Math.exp(tab[i]);
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>la puissance</u> d'un tableau.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab Le tableau.
   * @param n La puissance.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] pow(double[] tab, double n)
  {
    //----------------------------------------------
    // VERSION LABO : Les fonctions discrtes.
    //------------------------------------------------
    //throw new SNoImplementationException();
    
    double[] result = new double[tab.length];
    
    for(int i = 0; i < tab.length; i++)
      result[i] = Math.pow(tab[i], n);
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>le sinus</u> d'un tableau. Les valeurs du tableau seront considres en <b>radian</b>.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab Le tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] sin(double[] tab)
  {
    double[] result = new double[tab.length];
    
    for(int i = 0; i < tab.length; i++)
      result[i] = Math.sin(tab[i]);
    
    return result;
  }
  
  /**
   * Mthode pour raliser <u>le cosinus</u> d'un tableau. Les valeurs du tableau seront considres en <b>radian</b>.
   * Un nouveau tableau sera retourn contenant le rsultat de l'opration mathmatique.
   * 
   * @param tab Le tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] cos(double[] tab)
  {
    double[] result = new double[tab.length];
    
    for(int i = 0; i < tab.length; i++)
      result[i] = Math.cos(tab[i]);
    
    return result;
  }
  
  /**
   * Mthode pour raliser la fonction ReLU d'un tableau. 
   * 
   * @param tab Le tableau.
   * @return Le tableau contenant le rsultat de l'opration mathmatique.
   */
  public static double[] relu(double[] tab)
  {
    //-------------------------------------------------------
    // VERSION LABO : L'apprentissage par rseau de neurones.
    //-------------------------------------------------------
    //throw new SNoImplementationException();
    
    double[] result = new double[tab.length];
    
    for(int i = 0; i < tab.length; i++)
      result[i] = Math.max(0.0, tab[i]);
    
    return result;
  }
  
  /**
   * Mthode pour valuer la fonction softmax sur les composantes d'un tableau.
   * Le rsultat de la fonction sera enregistr dans le tableau result.
   * 
   * @param tab Le tableau des donnes sur lequel la fonction softmax sera effectue.
   * @param result Le rsultat de la fonction softmax.
   * @return Le tableau contenant les composantes du rsultat de la fonction softmax. 
   * @throws ArrayIndexOutOfBoundsException Si les deux tableaux n'ont pas la mme taille.
   */
  public static double[] softmax(double[] tab, double[] result) throws ArrayIndexOutOfBoundsException
  {
    // Vrification de la taille des tableaux.
    if(tab.length != result.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 015 : Le tableau n'a pas la mme taille que le tableau du rsultat (" + tab.length + " et " + result.length + ").");
      
    throw new SNoImplementationException();
    
  }
    
  /**
   * Mthode pour obtenir la <b>plus grande valeur</b> d'un tableau.
   * 
   * @param tab Le tableau.
   * @return La plus grande valeur du tableau.
   */
  public static int findMax(int[] tab)
  {
    int max = tab[0]; 
    
    for(int v : tab)
      if(v > max)
        max = v;
    
    return max;
  }
  
  /**
   * Mthode pour obtenir la <b>plus grande valeur</b> d'un tableau.
   * 
   * @param tab Le tableau.
   * @return La plus grande valeur du tableau.
   */
  public static double findMax(double[] tab)
  {
    double max = tab[0]; 
    
    for(double v : tab)
      if(v > max)
        max = v;
    
    return max;
  }
  
  /**
   * <p>
   * Mthode pour obtenir la <b>plus grande valeur sans tre zro</b> d'un tableau.
   * </p>
   * <p>
   * Cependant, si le <u>tableau contient uniquement des zros</u>, la valeur retourne sera zro.
   * </p>
   * 
   * @param tab Le tableau.
   * @return La plus grande valeurs (sauf zro) du tableau.
   */
  public static double findMaxNotZero(double[] tab)
  {
    double max = 0.0;
    int i = 0;
    
    // Obtenir le 1er chiffre pas gale  zro.
    while(max == 0.0 && i < tab.length)
    {
      if(tab[i] != 0.0)
        max = tab[i];
      
      i++;
    }
     
    // Vrifier si l'on a pas trouv de chiffre autre que zro.
    if(i == tab.length)
      return 0.0;
    
    // Chercher le chiffre le plus grand sauf pour zro.
    for(i = 0; i < tab.length; i++)
      if(tab[i] != 0.0)
        if(tab[i] > max)
          max = tab[i];
    
    return max;
  }
  
  /**
   * Mthode pour obtenir la <b>plus grande valeur absolue</b> d'un tableau.
   * 
   * @param tab Le tableau.
   * @return La plus grande valeur absolue du tableau.
   */
  public static double findAbsoluteMax(double[] tab)
  {
    double max = Math.abs(tab[0]); 
    
    for(double v : tab)
      if(Math.abs(v) > max)
        max = Math.abs(v);
    
    return max;
  }
  
  /**
   * Mthode pour obtenir la plus grande valeur d'un tableau en explorant uniquement des indices du tableau autoris.
   * Les indices autoriss sont dtermin par le second tableau o 1.0 signifie que l'indice peut tre autorise et 0.0 n'aurorisant pas.
   * Si aucun indice est autoris, une exception sera lance.
   * 
   * @param array
   * @param authorization
   * @throws IllegalArgumentException Si aucun indice est autoris dans la recherche.
   * @throws ArrayIndexOutOfBoundsException Si les deux tableaux n'ont pas la mme taille.
   * @return La valeur maximale autorise dans le tableau.
   */
  public static double findMaxInAllowedIndex(double[] array, boolean[] authorization) throws IllegalArgumentException, ArrayIndexOutOfBoundsException
  {
    if(array.length != authorization.length)
      throw new ArrayIndexOutOfBoundsException("Erreur SArrays 009 : Les deux tableaux n'ont pas la mme longueur (" + array.length + " et " + authorization + ").");
    
    double value = Double.NEGATIVE_INFINITY;
    
    for(int i = 0; i < array.length; i++)
      if(authorization[i])
        if(array[i] > value)
          value = array[i];
    
    // Vrification s'il n'y a pas eu d'autorisation.
    if(Double.isInfinite(value))
      throw new IllegalArgumentException("Error SArrays 010 : Il n'y a pas eu de recherche dans le tableau " + Arrays.toString(array) + " dont l'autorisation est " + Arrays.toString(authorization));
    
    return value;
  }
  
  /**
   * Mthode pour obtenir l'indice dans le tableau de la <b>valeur la plus grande</b>.
   * 
   * @param tab Le tableau.
   * @return L'indice dans le tableau o l'on trouve la valeur la plus grande.
   */
  public static int findIndexOfMax(double[] tab)
  {
    // Choisir la meilleure cote.
    int index = 0;
    double value = tab[0];
    
    // Parcourir le tableau  la recherche de la plus grande valeur.
    for(int i = 1; i < tab.length; i++)
      if(tab[i] > value)
      {
        value = tab[i];
        index = i;
      }
    
    return index;
  }
  
  /**
   * Mthode pour obtenir l'indice dans le tableau de la <b>valeur absolue la plus grande</b>.
   * 
   * @param tab Le tableau.
   * @return L'indice dans le tableau o l'on trouve la valeur absolue la plus grande.
   */
  public static int findIndexOfAbsoluteMax(double[] tab)
  {
    // Choisir la meilleure cote.
    int index = 0;
    double value = Math.abs(tab[0]);
    
    // Parcourir le tableau  la recherche de la plus grande valeur.
    for(int i = 1; i < tab.length; i++)
      if(Math.abs(tab[i]) > value)
      {
        value = Math.abs(tab[i]);
        index = i;
      }
    
    return index;
  }
  
  /**
   * Mthode pour obtenir la <b>plus petite valeur</b> d'un tableau.
   * 
   * @param tab Le tableau.
   * @return La plus petite valeur du tableau.
   */
  public static int findMin(int[] tab)
  {
    int min = tab[0]; 
    
    for(int v : tab)
      if(v < min)
        min = v;
    
    return min;
  }
  
  /**
   * Mthode pour obtenir la <b>plus petite valeur</b> d'un tableau.
   * 
   * @param tab Le tableau.
   * @return La plus petite valeur du tableau.
   */
  public static double findMin(double[] tab)
  {
    double min = tab[0]; 
    
    for(double v : tab)
      if(v < min)
        min = v;
    
    return min;
  }
  
  /**
   * Mthode pour convertir un tableau d'entier vers un tableau de nombre rel entre une valeur minimale et maximale.
   * La correspondance entre les valeurs converties sera linaire.
   * 
   * @param data Les entiers  convertir.
   * @param min La borne minimale.
   * @param max La borne maximale.
   * @return Le tableau converti.
   * @throws SRuntimeException Si les bornes sont mal dfinies.
   */
  public static double[] mappingIntToDouble(int[] data, double min, double max) throws SRuntimeException
  {
    if(min > max)
      throw new SRuntimeException("Error SArrays 011 : La borne minimale " + min + " et la borne maximale " + max + " sont mal dfinies.");
    
    int min_value = findMin(data);
    int max_value = findMax(data);
    
    double[] result = new double[data.length];
    
    // Itrer sur l'ensemble des lments du tableau.
    for(int i = 0; i < result.length; i++)
    {
      double t = SMath.reverseLinearInterpolation((double)data[i], (double)min_value, (double)max_value);
      result[i] = SMath.linearInterpolation(min, max, t);
    }
    
    return result;
  }
  
  /**
   * Mthode pour convertir un tableau de nombre rel vers un tableau d'entier entre une valeur minimale et maximale.
   * La correspondance entre les valeurs converties sera linaire.
   * 
   * @param data Les donnes  convertir.
   * @param min La borne minimale.
   * @param max La borne maximale.
   * @return Le tableau converti.
   * @throws IllegalArgumentException Si les bornes sont mal dfinies.
   */
  public static int[] mappingDoubleToInt(double[] data, int min, int max) throws IllegalArgumentException
  {
    return mappingDoubleToInt(data, findMin(data), findMax(data), min, max);
  }
  
  /**
   * ...
   * 
   * @param data
   * @param double_min
   * @param double_max
   * @param int_min
   * @param int_max
   * @return
   * @throws IllegalArgumentException
   */
  public static int[] mappingDoubleToInt(double[] data, double double_min, double double_max, int int_min, int int_max) throws IllegalArgumentException
  {
    if(int_min > int_max)
      throw new IllegalArgumentException("Error SArrays 012 : La borne entire minimale " + int_min + " et la borne entire maximale " + int_max + " sont mal dfinies.");
    
    if(double_min > double_max)
      throw new IllegalArgumentException("Error SArrays 013 : La borne double minimale " + double_min + " et la borne double maximale " + double_max + " sont mal dfinies.");
       
    int[] result = new int[data.length];
    
    // Itrer sur l'ensemble des lments du tableau.
    for(int i = 0; i < result.length; i++)
    {
      double t = SMath.reverseLinearInterpolation(data[i], double_min, double_max);
      result[i] = (int)SMath.linearInterpolation(int_min, int_max, t);
    }
    
    return result;
  }
  
  /**
   * <p>
   * Mthode permettant de gnrer un tableau contenant les lments identique de deux tableaux de valeur <ul>pralablement tri</ul>.
   * </p>
   * 
   * <p>
   * <b>REMARQUE</b> : Le fonctionnement de cette mthode ne sera pas valide si les deux tableaux passs en paramtre ne sont pas palablement tri.
   * </p>
   * Rfrence : https://stackoverflow.com/questions/32676381/find-intersection-of-two-arrays
   * 
   * @param tab1 Le 1ier tableau  comparer.
   * @param tab2 Le 2ime tableau  comparer.
   * @return Un tableau comprenant les lments identiques de deux tableaux (l'intersection des deux tableaux).
   */
  public static double[] intersectionSortedArray(double[] tab1, double[] tab2)
  {
    return intersectionSortedArray(tab1, tab2, SMath.EPSILON);
  }
  
  /**
   * <p>
   * Mthode permettant de gnrer un tableau contenant les lments identique de deux tableaux de valeur <ul>pralablement tri</ul>.
   * </p>
   * 
   * <p>
   * <b>REMARQUE</b> : Le fonctionnement de cette mthode ne sera pas valide si les deux tableaux passs en paramtre ne sont pas palablement tri.
   * </p>
   * Rfrence : https://stackoverflow.com/questions/32676381/find-intersection-of-two-arrays
   * 
   * @param tab1 Le 1ier tableau  comparer.
   * @param tab2 Le 2ime tableau  comparer.
   * @param epsilon La prcision de la comparaison.
   * @return Un tableau comprenant les lments identiques de deux tableaux (l'intersection des deux tableaux).
   */
  public static double[] intersectionSortedArray(double[] tab1, double[] tab2, double epsilon)
  {
    double intersection[] = new double[Math.min(tab1.length, tab2.length)];
    int count = 0;

    int i = 0; int j = 0;
    while (i < tab1.length && j < tab2.length)
    {
      // Vrifier s'il y a galit.
      if(SMath.nearlyEquals(tab1[i], tab2[j], epsilon)) 
      {
        intersection[count] = tab1[i];
        count++;
        i++;
        j++;
      }
      else
        // Avancer dans la recherche des lments.
        if (tab1[i] < tab2[j]) 
          i++;                    // avancer dans le tableau 1
        else
          j++;                    // avancer dans le tableau 2
    }

    // Construire un tableau avec l'espace mmoire minimum.
    double[] result = new double[count];
    for(int k = 0; k < result.length; k++)
      result[k] = intersection[k];
    
    return result;
  }
  
  /**
   * Mthode pour changer la dimension d'un tableau. Si la nouvelle taille est infrieure  la prcdente, les derniers lments ne seront pas copi.
   * Si la nouvelle taille est suprieure  la prcdente, des valeurs nulles sont attribues aux positions suprieures. 
   * 
   * @param array Le tableau  copier.
   * @param new_size La nouvelle dimension du tableau.
   * @return Un tableau avec une copie des donnes avec une taille diffrente.
   */
  public static double[] resize(double[] array, int new_size)
  {
      // Remplir un nouveau tableau ayant seulement la taille demand.
      // Tous les lments du tableau prcdent 
      double[] result = new double[new_size];
      
      for(int i = 0; i < new_size; i++)
        result[i] = array[i];
     
      return result;
  }
  
  /**
   * Mthode pour changer la dimension d'un tableau. Si la nouvelle taille est infrieure  la prcdente, les derniers lments ne seront pas copi.
   * Si la nouvelle taille est suprieure  la prcdente, des valeurs nulles sont attribues aux positions suprieures. 
   * 
   * @param array Le tableau  copier.
   * @param new_size La nouvelle dimension du tableau.
   * @return Un tableau avec une copie des donnes avec une taille diffrente.
   */
  public static int[] resize(int[] array, int new_size)
  {
      // Remplir un nouveau tableau ayant seulement la taille demand.
      // Tous les lments du tableau prcdent 
      int[] result = new int[new_size];
      
      for(int i = 0; i < new_size; i++)
        result[i] = array[i];
     
      return result;
  }
  
  /**
   * Mthode pour faire l'criture d'un tableau dans un fichier texte.
   * 
   * @param file_name Le nom du fichier.
   * @param values Le tableau des valeurs  crire.
   */
  public static void write(String file_name, double[] values)
  {
    try{
      
      FileWriter fw = new FileWriter(file_name);
      BufferedWriter bw = new BufferedWriter(fw);
    
      // Faire l'criture de chaque valeur et changer de ligne.
      for(int i = 0; i < values.length; i++)
      {
        bw.write(Double.toString(values[i]));
        bw.newLine();
        bw.flush();
      }
      
      bw.close(); //  fermer celui-ci en premier, sinon, ERROR !!!
      fw.close();   
      
    }catch(IOException ioe){
      SLog.logWriteLine("Message SArrays - Une erreur de type I/O est survenue.");
      ioe.printStackTrace();
    }
  }
  
  /**
   * Mthode pour convertir les donnes d'un tableau de donne en un tableau de probabilit.
   * Les donnes seront <u>traites en valeur absolue</u>.
   * Si le tableau contient uniquement des zros, il restera avec des zros et la probablitit totale du tableau sera zro (et non 1.0).
   * 
   * @param array Le tableau  convertir en probabilit.
   */
  public static void convertToAbsoluteProbability(double[] array) throws IllegalArgumentException
  {
    // Pour calculer la somme des lments du tableau.
    double sum = 0.0;
    
    // Faire la somme du tableau et vrifier qu'il n'y ait pas de valeur ngative.
    for(int i = 0; i < array.length; i++)
      sum += Math.abs(array[i]);
    
    // Faire la normalisation en probabilit.
    if(sum == 0.0)
      throw new IllegalArgumentException("Error SArrays 014 : Le tableau suivant ne peut pas tre convertie en probabilit : " + Arrays.toString(array));
    
    for(int i = 0; i < array.length; i++)
      array[i] = Math.abs(array[i]) / sum;
  }
    
  /**
   * Mthode pour faire la somme de tous les lments d'un tableau.
   * 
   * @param array Le tableau.
   * @return La somme de tous les lments du tableau.
   */
  public static double sumElementsArray(double[] array)
  {
    double v = 0.0;
    
    for(int i = 0; i < array.length; i++)
      v += array[i];
    
    return v;
  }
  
  /**
   * Mthode pour fractionner une valeur dans un tableau. La valeur sera la mme dans chaque lment du tableau
   * et la somme des lments sera gal  la valeur  fractionner.
   * 
   * @param value La valeur  fractionner.
   * @param array_size La taille du tableau.
   * @return Le tableau de valeur fractionner.
   * @throws IllegalArgumentException Si le tableau a une taille infrieur  1.
   */
  public static double[] splitValueInArray(double value, int array_size) throws IllegalArgumentException
  {
    if(array_size < 1)
      throw new IllegalArgumentException("Erreur SArrays 15 : Le tableau de fractionnement de valeur ne peut pas avoir une taille infrieur  1. (array_size = " + array_size + ").");
    
    double[] array = new double[array_size];
    
    double v = value / (double) array_size;
    
    for(int i = 0 ; i < array_size; i++)
      array[i] = v;
    
    return array;
  }
  
  /**
   * Mthode pour mettre bout  bout des tableaux de type double.
   * 
   * Rfrence : https://www.techiedelight.com/merge-multiple-arrays-java/
   * 
   * @param arrays Les tableaux.
   * @return Le tableau contenant la fusion des tableaux.
   */
  public static double[] merge(double[]... arrays)
  {
    int finalLength = 0;
    for (double[] array : arrays) {
      finalLength += array.length;
    }

    double[] dest = null;
    int destPos = 0;

    for (double[] array : arrays)
    {
      if (dest == null) {
        dest = Arrays.copyOf(array, finalLength);
        destPos = array.length;
      } else {
        System.arraycopy(array, 0, dest, destPos, array.length);
        destPos += array.length;
      }
    }
    
    return dest;
  }
  
}
